/*
Variable width font drawing library for DS (and GBA)

Copyright 2007 Damian Yerrick <pinoandchester@pineight.com>

This work is provided 'as-is', without any express or implied
warranty.  In no event will the authors be held liable for any
damages arising from the use of this work.

Permission is granted to anyone to use this work for any purpose,
including commercial applications, and to alter it and redistribute
it freely, subject to the following restrictions:

1. The origin of this work must not be misrepresented; you must
   not claim that you wrote the original work. If you use
   this work in a product, an acknowledgment in the product
   documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must
   not be misrepresented as being the original work.
3. This notice may not be removed or altered from any source
   distribution.

"Source" is the preferred form of a work for making changes to it.

*/

#include <string.h>
#include <stdio.h>
#include "fontdraw.h"

typedef u16 NAMETABLE[32][32];
#define MAP ((NAMETABLE *)BG_MAP_RAM(0))
#define MAP_SUB ((NAMETABLE *)BG_MAP_RAM_SUB(0))


extern const unsigned char vwfont_bin[];

typedef struct GlyphRec {
  unsigned short dataOffset;
  unsigned char glyphWidth;
  unsigned char reserved;
} GlyphRec;

typedef struct GlyphRange {
  unsigned short baseCodePoint;
  unsigned char nCodePoints;
  unsigned char reserved1;
  unsigned short baseGlyphRec;
  unsigned short reserved2;
} GlyphRange;


unsigned int fontdraw_charWidth(int glyph) {
  glyph &= 0xFF;
  if (glyph < vwfont_bin[1]) {
    return 0;
  }
  glyph -= vwfont_bin[1];
  if (vwfont_bin[2] != 0 && glyph >= vwfont_bin[2]) {
    return 0;
  }
  const GlyphRec *glyphRec = 
    ((const GlyphRec *)(vwfont_bin + vwfont_bin[0])) + glyph;
  return glyphRec->glyphWidth;
}

unsigned int fontdraw_strWidth(const char *s) {
  unsigned int width = 0;
  for (; *s; ++s) {
    width += fontdraw_charWidth(*s);
  }
  return width;
}

size_t fontdraw_cutStr(const char *s, int targetWidth) {
  size_t len;
  for (len = 0; s[len]; ++len) {
    int charWidth = fontdraw_charWidth(s[len]);
    if (charWidth > targetWidth) {
      return len;
    }
    targetWidth -= charWidth;
  }
  return len;
}

static unsigned int fontdraw_putchar(u32 *dst, unsigned int colStride, int wid, int x, int glyph) {
  glyph &= 0xFF;
  if (glyph < vwfont_bin[1]) {
    return 0;
  }
  glyph -= vwfont_bin[1];
  if (vwfont_bin[2] != 0 && glyph >= vwfont_bin[2]) {
    return 0;
  }
  const GlyphRec *glyphRec = 
    ((const GlyphRec *)(vwfont_bin + vwfont_bin[0])) + glyph;
  const unsigned char *data = vwfont_bin + glyphRec->dataOffset;
  unsigned int dataShift = 2;
  unsigned int dataBits = *data++;
  unsigned int pixelCode = dataBits & 0x03;

  // Convert x to tile column address and bit address within tile
  dst += colStride * (x >> 3);
  x = (x & 0x07) << 2;
  
  if (dataShift >= 8) {
    dataShift = 2;
    dataBits = *data++;
  }

  for (unsigned int height = vwfont_bin[3];
       height > 0;
       --height, ++dst) {
    int eol = 0;
    int xLine = x;
    int widLeft = wid;
    u32 *dstLine = dst;
    u32 dstBits = *dstLine;
    while (!eol) {

      // Process a pixel instruction
      if (pixelCode == 0) {
        eol = 1;
      } else if (widLeft > 0) {

        // Change pixel and move to next pixel
        if (pixelCode > 1) {
          //dstBits &= ~(0x0F << xLine);
          dstBits |= (pixelCode - 1) << xLine;
        }
        xLine += 4;
        --widLeft;
      }

      // If finished with this tile, write back changed bits
      if (xLine >= 32) {
        xLine = 0;
        *dstLine = dstBits;
        dstLine += colStride;
        dstBits = *dstLine;
      }
      
      // Decode next pixel instruction
      if (dataShift >= 8) {
        dataShift = 0;
        dataBits = *data++;
      }

      // Decode a byte into pixel instructions
      pixelCode = (dataBits >> dataShift) & 0x03;
      dataShift += 2;
    }

    // Write back changed bits
    if (xLine > 0) {
      *dstLine = dstBits;
    }
  }
  return glyphRec->glyphWidth;
}

static unsigned int fontdraw_putline(u32 *dst, unsigned int colStride, int wid, int x, const char *src) {
  unsigned int startX = x;
  for (int c = *src; c != 0 && wid > 0; c = *++src) {
    int chWidth = fontdraw_putchar(dst, colStride, wid, x, c & 0xFF);
    x += chWidth;
    wid -= chWidth;
  }
  return x - startX;
}

static unsigned int fontdraw_putchars(u32 *dst, unsigned int colStride, int wid, int x, const char *src, int len) {
  unsigned int startX = x;
  for (int c = *src;
       len > 0;
       --len, c = *++src) {
    int chWidth = fontdraw_putchar(dst, colStride, wid, x, c & 0xFF);
    x += chWidth;
    wid -= chWidth;
  }
  return x - startX;
}

const VWFWindow vwfTop = {
  .left = 0, .top = 0, .width = 32, .height = 24,
  .chrBase = (u32 *)BG_TILE_RAM(0),
  .map = 31,
  .core = 0,
  .mapTileBase = 0
};

const VWFWindow vwfTouch = {
  .left = 0, .top = 0, .width = 32, .height = 24,
  .chrBase = (u32 *)BG_TILE_RAM_SUB(0),
  .map = 31,
  .core = 1,
  .mapTileBase = 0
};

static
void vwfRectfillColumn(u32 *dst, unsigned int colStride,
                       unsigned int l, unsigned int t,
                       unsigned int r, unsigned int b,
                       unsigned int c)
{
  u32 mask = 0xffffffffU << (4 * l);
  mask &= 0xffffffffU >> (4 * (8 - r));
  c &= mask;
  mask = ~mask;

  for (; t < b; t++) {
    dst[t] = (dst[t] & mask) | c;
  }
}

void vwfRectfill(const VWFWindow *v, int l, int t, int r, int b, int c)
{
  u32 *dst = v->chrBase;
  c &= 0x0000000F;
  c |= c << 4;
  c |= c << 8;
  c |= c << 16;
  
  unsigned int x = l;
  unsigned int stride = v->height * 8;
  u32 *tile = dst + stride * (l >> 3);
  
  if (t < 0) {
    t = 0;
  }
  if (b > v->height * 8) {
    b = v->height * 8;
  }

  for(x = l; x < r; x = (x + 8) & -8) {
    vwfRectfillColumn(tile, stride, x & 7, t,
                      (r & -8) > (x & -8) ? 8 : (r & 7),
                      b, c);
    tile += stride;
  }
}

void vwfHline(const VWFWindow *v, int l, int t, int r, int c) {
  if (r < l) {
    int temp = l;
    l = r;
    r = temp;
  }
  vwfRectfill(v, l, t, r, t + 1, c);
}

void vwfVline(const VWFWindow *v, int l, int t, int b, int c) {
  if (b < t) {
    int temp = t;
    t = b;
    b = temp;
  }
  vwfRectfill(v, l, t, l + 1, b, c);
}

void vwfRect(const VWFWindow *v, int l, int t, int r, int b, int c) {
  vwfVline(v, l, t, b, c);
  vwfVline(v, r - 1, t, b, c);
  vwfHline(v, l, t, r, c);
  vwfHline(v, l, b - 1, r, c);
}

void vwfWinClear(const VWFWindow *w) {
  size_t nBytes = 32 * w->width * w->height;
  memset(w->chrBase, 0, nBytes);
}

void vwfWinInit(const VWFWindow *w) {
  NAMETABLE *dst = w->core ? &(MAP_SUB[w->map]) : &(MAP[w->map]);

  int mapTile = w->mapTileBase;
  for (int x = w->left; x < w->left + w->width; ++x) {
    for (int y = w->top; y < w->top + w->height; ++y) {
      (*dst)[y][x] = mapTile++;
    }
  }
  vwfWinClear(w);
}

unsigned int vwfPuts(const VWFWindow *w,
             const char *str,
             int x, int y) {
  return fontdraw_putline(w->chrBase + y, 8 * w->height,
                           w->width * 8 - x, x, str);
}

unsigned int vwfPutc(const VWFWindow *w,
             int c,
             int x, int y) {
  int width = 8 * w->width - x;
  if (width >= fontdraw_charWidth(c)) {
    return fontdraw_putchar(w->chrBase + y, 8 * w->height,
                             w->width * 8 - x, x, c & 0xFF);
  } else {
    return 0;
  }
}

// old api emulation

#if 0
void fontdraw_demo_putline(int x, int y, const char *str) {
  const VWFWindow *w = &vwfTop;
  if (y >= 288) {
    w = &vwfTouch;
    y -= 288;
  }
  vwfPuts(w, str, x, y);
}
#endif

void fontdraw_setupVRAM(int sub) {
  const VWFWindow *w = sub ? &vwfTouch : &vwfTop;

  if (sub == 0) {
    BGCTRL[0] = BG_MAP_BASE(31) | BG_TILE_BASE(0);
    BG_OFFSET[0].x = 0;
    BG_OFFSET[0].y = 0;
  } else {
    BGCTRL_SUB[0] = BG_MAP_BASE(31) | BG_TILE_BASE(0);
    BG_OFFSET_SUB[0].x = 0;
    BG_OFFSET_SUB[0].y = 0;
  }
  
  vwfWinInit(w);
}

void fontdraw_cls(int sub) {
  vwfWinClear(sub ? &vwfTouch : &vwfTop);
}
