/*
memtest.c, the engine of a RAM tester
based on the techniques of memtest86

Copyright 2007 Damian Yerrick

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

*/

#include "memtest.h"
#include <stdlib.h>

#ifdef TEST_SUITE
#include <stdio.h>  

enum {
  /* size of canary to make sure that loops don't overflow buffers */
  PC_ALLOC_OVERHEAD = 4
};
static void dump(volatile const void *const start, const size_t length) {
  volatile const unsigned char *bytes = start;
  unsigned int bytesThisLine = 0;
  size_t i;

  for (i = 0; i < length + PC_ALLOC_OVERHEAD; ++i) {
    printf("%02x ", bytes[i]);
    if (++bytesThisLine >= 24) {
      bytesThisLine = 0;
      fputc('\n', stdout);
    }
  }
  if (bytesThisLine) {
    fputc('\n', stdout);
  }
}
#else
static inline void dump(volatile const void *const start, const size_t length) {
  (void)start;
  (void)length;
}
#endif

volatile void *movingInversions(volatile void *const start, const size_t length,
                                const unsigned int pattern) {
  volatile memtest_word_t *here;
  size_t wordsLeft;

  /* Step 1: Clear memory */
  for (here = (volatile memtest_word_t *)start,
       wordsLeft = length / sizeof(here[0]);
       wordsLeft > 0;
       ++here, --wordsLeft) {
    *here = (memtest_word_t)pattern;
  }
  dump(start, length);

  /* Step 2: Invert forwards */
  for (here = (volatile memtest_word_t *)start,
       wordsLeft = length / sizeof(here[0]);
       wordsLeft > 0;
       ++here, --wordsLeft) {
    if (*here != (memtest_word_t)pattern) {
      return here;
    }
    *here = (memtest_word_t)~pattern;
  }
  dump(start, length);

  /* Step 2: Invert backwards */
  for (wordsLeft = length / sizeof(here[0]),
       here = (volatile memtest_word_t *)start + (wordsLeft - 1);
       wordsLeft > 0;
       --here, --wordsLeft) {
    if (*here != (memtest_word_t)~pattern) {
      return here;
    }
    *here = (memtest_word_t)pattern;
  }
  dump(start, length);
  
  /* Success */
  return NULL;
}

volatile void *slideTest(volatile void *const start, const size_t length,
                         const unsigned int pattern, const size_t phase) {
  size_t curPhase;
  volatile memtest_word_t *here;
  size_t wordsLeft;

  /* Step 1: Write pattern */
  for (here = (volatile memtest_word_t *)start,
       wordsLeft = length / sizeof(here[0]),
       curPhase = phase % MEMTEST_SLIDE;
       wordsLeft > 0;
       ++here, --wordsLeft, --curPhase) {
    unsigned int value;

    if (curPhase == 0) {
      curPhase = MEMTEST_SLIDE;
      value = (memtest_word_t)~pattern;
    } else {
      value = (memtest_word_t)pattern;
    }

    *here = value;
  }
  dump(start, length);

  /* Step 2: Read back pattern */  
  for (here = (volatile memtest_word_t *)start,
       wordsLeft = length / sizeof(here[0]),
       curPhase = phase % MEMTEST_SLIDE;
       wordsLeft > 0;
       ++here, --wordsLeft, --curPhase) {
    unsigned int value;

    if (curPhase == 0) {
      curPhase = MEMTEST_SLIDE;
      value = (memtest_word_t)~pattern;
    } else {
      value = (memtest_word_t)pattern;
    }

    if (*here != value) {
      return here;
    }
  }
  
  /* Success */
  return NULL;
}

volatile void *fullSlideTest(volatile void *const start, const size_t length,
                             const unsigned int pattern) {
  size_t phase;
  
  for (phase = 0; phase < MEMTEST_SLIDE; ++phase) {
    volatile void *failure = slideTest(start, length, pattern, phase);
    if (failure) {
      return failure;
    }
  }
  return NULL;
}

#ifdef TEST_SUITE

enum {
  TEST_SIZE = 128,
  AS_IF_PATTERN = 0x1FA5
};

int main(void) {
  void *data;
  volatile void *failure;

  printf("allocating %d bytes ", TEST_SIZE);
  data = malloc(PC_ALLOC_OVERHEAD + TEST_SIZE);
  if(!data) {
    fputs("failed!\n", stdout);
    return EXIT_FAILURE;
  }
  printf("at 0x%p\n", data);
  
  printf("testing moving inversions\n");
  failure = movingInversions(data, TEST_SIZE, AS_IF_PATTERN);
  if (failure) {
    fputs("freeing memory\n", stdout);
    free(data);
    printf("test failed at address 0x%p\n", failure);
    return EXIT_FAILURE;
  }

  printf("testing slide test\n");
  failure = fullSlideTest(data, TEST_SIZE, AS_IF_PATTERN);
  if (failure) {
    fputs("freeing memory\n", stdout);
    free(data);
    printf("test failed at address 0x%p\n", failure);
    return EXIT_FAILURE;
  }
  
  fputs("freeing memory\n", stdout);
  free(data);
  fputs("success.\n", stdout);
  return EXIT_SUCCESS;
}

#endif  /* end of PC based test suite */
