/*
header for memtest.c, the engine of a RAM tester
based on the techniques of memtest86

Copyright 2007 Damian Yerrick

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

*/

/* Rationale for above license

For short programs that test a given library, I try to use the same
license as the library.  This program was originally designed to test
Lick's RAM Unlocking API, which was relicensed as MIT on 2007-09-06.

*/

#ifndef P8_MEMTEST_H
#define P8_MEMTEST_H

#include <sys/types.h>

/**
 * This module performs the "moving inversions" and "modulo X" tests
 * described at <http://www.memtest86.com/tech.html#philo>.
 * In this program, "modulo X" is called "the slide test".
 */
 
/**
 * First a way to perform assertions at compile time when
 * #ifdef is too early and if() is too late, such as tests for
 * sizeof.
 * @param name the name of this assertion; the error message
 * will use this name
 * @param condition if this equal zero, the compiler will issue a
 * diagnostic and fail to compile this translation unit
 */
#define COMPILE_TIME_ASSERT(name, condition) \
  extern unsigned char name[(condition) ? 1 : -1]

COMPILE_TIME_ASSERT(testing_cta, 1);
/* COMPILE_TIME_ASSERT(cta_failure, 0); */

/**
 * The size of a memory word being tested.  This should be as wide
 * as a hardware memory cell.
 */
typedef unsigned short int memtest_word_t;

/**
 * For time efficiency and argument handling predictability,
 * we require that the test patterns fit in an unsigned int.
 */
COMPILE_TIME_ASSERT(slideTest_int_vs_memtest_word_t,
                    sizeof(memtest_word_t) <= sizeof(unsigned int));
/**
 * Distance between complemented bytes in memtest86 is 20 units.
 * The rationale page http://www.memtest86.com/tech.html#philo
 * appears to imply that any non-power-of-two number would work,
 * and that 20 is a tradeoff between speed and thoroughness.
 * I use a prime number here for extra safety.
 */
enum {
  MEMTEST_SLIDE = 19
};

/**
 * Performs the moving inversion testfor each value of phase between
 * 0 and MEMTEST_SLIDE - 1.
 * @param start the starting address of the block
 * @param length the length of the block
 *   (in sizeof() units, which are chars)
 * @param pattern the bit pattern to use
 * @return the first location that fails, or NULL if successful
 */
volatile void *movingInversions(volatile void *const start, const size_t length,
                                const unsigned int pattern);
/**
 * Performs the slide test on a block of memory
 * @param start the starting address of the block
 * @param length the length of the block
 *   (in sizeof() units, which are chars)
 * @param pattern the bit pattern to use
 * @param phase the starting offset to complement
 *   (between 0 and MEMTEST_SLIDE - 1; out-of-bounds
 *   values will be wrapped to fit)
 * @return the first location that fails, or NULL if successful
 */
volatile void *slideTest(volatile void *const start, const size_t length,
                         const unsigned int pattern, const size_t phase);
                         
/**
 * Performs the slide test for each value of phase between
 * 0 and MEMTEST_SLIDE - 1.
 * @param start the starting address of the block
 * @param length the length of the block
 *   (in sizeof() units, which are chars)
 * @param pattern the bit pattern to use
 * @return the first location that fails, or NULL if successful
 */
volatile void *fullSlideTest(volatile void *const start, const size_t length,
                             const unsigned int pattern);

#endif
